import yaml
import pandas as pd
import os
from Bio.SeqRecord import SeqRecord
from mooda.evolutionary_algorithm import *
from pathlib import Path


class MoodaFile:


    def __init__(self, yaml_file):
        self.yaml_file = yaml_file


    def set_yaml_file(self, config):
        self.yaml_file = config

    def __get_path(self, output_dir):
        self.path = Path(output_dir)

    def __check_dir(self, output_dir):
        test_dir = os.path.isdir(output_dir)
        if not test_dir:
            os.makedirs(output_dir)




    def __file_check(self, output_dir, suffix):
        n = 1

        wildcard = self.path.parts[-1]
        filename = wildcard + suffix
        self.__check_dir(output_dir)
        while filename in os.listdir(output_dir):
            n += 1
            filename = (
                str(wildcard)
                + "_"
                + str(n)
                + suffix
            )
        filename = os.path.join(output_dir, filename)
        return filename

    def save_log(self, options, ag):
        suffix = "_logfile.yaml"
        self.__check_dir(options.output_dir)
        filename = self.path.parts[-1] + suffix
        output = os.path.join(options.output_dir,filename)
        parameters = dict()
        parameters["max_iterations"] = ag.max_iterations

        log = dict(
            {
                "Input file": options.input_file,
                "Outputs": options.output_dir,
                "Objective functions": ag.yaml["Algorithm"]["objective_functions"],
                "Operators": ag.yaml["Algorithm"]["operators"],
                "Max iterations": ag.max_iterations,
                "Output genbank and fasta": options.sequence_output,
                "GenBank data": str(ag.gebank_record),
                "Sequence": str(
                    ag.gebank_record.seq[0:100] + "...." + ag.gebank_record.seq[-100:]
                ),
                "start": str(ag.time_start),
                "time process": str(ag.time_end - ag.time_start),
                "end": str(ag .time_end),
            }
        )
        if options.algorithm == "mc":
            log["Algorithm"] = "Montecarlo"
        elif options.algorithm == "mo":
            log["Algorithm"] = "Multi-Objective"
            log["Population"] = ag.population.size
            log["Archive size"] = ag.archive.length
        yaml.dump(
            log, open(output, "w"), default_flow_style=False, allow_unicode=True
        )

    def save_mooda_result(self, options, ag):
        suffix = "_mooda_result.csv"
        self.__check_dir(options.output_dir)
        filename = self.path.parts[-1] + suffix
        output = os.path.join(options.output_dir, filename)
        objective_list = []
        for objective_function in ag.objective_functions:
            label = objective_function.__repr__()
            objective_list.append(label)
        fieldnames = ["Crowding distance"] + objective_list + ["Fragments", "Record"]

        individual_list = None
        if options.algorithm == "mo":
            individual_list = ag.population.individuals
        elif options.algorithm == "mc":
            individual_list = ag.archive.individuals

        dict_list = []
        for individual in individual_list:
            individual.objectives = [round(objective, 2) for objective in individual.objectives]
            objective_dictionary = dict(zip(objective_list, individual.objectives))
            objective_dictionary["Crowding distance"] = round(individual.crowding_distance, 2)
            objective_dictionary["Fragments"] = individual.fasta
            objective_dictionary["Record"] = individual.genbank
            dict_list.append(objective_dictionary)
        table = pd.DataFrame(dict_list, columns=fieldnames)
        table.to_csv(output)

    def save_fasta(self, options,ag):
        suffix = ".fasta"
        path_dir = Path(options.output_dir)
        wildcard = path_dir.parts[-1]
        if ag.outputfasta_gebank:
            individual_list =[]
            if options.algorithm == "mo":
                individual_list = ag.population.individuals
            elif options.algorithm == "mc":
                individual_list = ag.archive.individuals
            for ind in individual_list:
                ind.fasta = self.__file_check(options.output_dir, suffix)
                with open(ind.fasta, "w") as fastaseq:
                    for block in ind.blocks:
                        seq_record = SeqRecord(
                            ind.sequence[block[0] : block[1]],
                            id=ind.gebank_record.id,
                            name=ind.gebank_record.name,
                            description=">" + str(wildcard) + " " + str(block) + "\n",
                        )
                        SeqIO.write(seq_record, fastaseq, "fasta")

    def save_genbank(self, options, ag):
        suffix = ".genbank"
        if ag.outputfasta_gebank:
            individual_list = []
            if options.algorithm == "mo":
                individual_list = ag.population.individuals
            elif options.algorithm == "mc":
                individual_list = ag.archive.individuals
            for ind in individual_list:
                ind.genbank = self.__file_check(options.output_dir, suffix)
                record = SeqRecord(
                    ind.sequence,
                    id=ind.gebank_record.id,
                    name=ind.gebank_record.name,
                    description=ind.gebank_record.id,
                )
                for feature in ind.features:
                    record.features.append(feature)
                output_file = open(ind.genbank, "w")
                SeqIO.write(record, output_file, "genbank")

    def save(self,options, ag):
        self.__get_path(options.output_dir)
        self.save_fasta(options, ag)
        self.save_genbank(options, ag)
        self.save_log(options, ag )
        self.save_mooda_result(options, ag)




class Assembly:
    def __init__(self, yaml):
        self.yaml = yaml

    def apply(self, ind):
        pass


class Gibson(Assembly):
    def __set_junction_size(self):
        self.junction_size = self.yaml["Algorithm"]["assemblies"][
            "mooda.assembly.Gibson"
        ]["junction_size"]

    def initialise(self):
        self.__set_junction_size()

    """
    Takes as input all blocks if there is no overlap it
    makes it
    """

    def apply(self, population):
        # for each block
        for ind in population.individuals:
            for block_index in range(len(ind.blocks[:-1])):
                next_block = block_index + 1
                # if there is no overlap make it
                if ind.blocks[block_index][1] == ind.blocks[next_block][0]:
                    ind.blocks[block_index][1] = ind.blocks[block_index][1] + self.junction_size



import numpy as np
import random
import math
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from mooda.config import *

"""
    Class Operator

    desc
"""


class Operator:
    def __init__(self, yaml):
        self.yaml = yaml


    def apply(self, ind):
        pass


"""
Class Block initialisator
"""


class BlockInitialiserOperator:
    """
    This method is used during the initialisation to create block : min_size< size<max_size
    """

    def __init__(self):
        super().__init__()
        self.max_block_size = None
        self.min_block_size = None
        self.step_size = None
        self.junction_size = None

    def _get_block_maxsize(self, ind):
        max_length_block = 0
        for block in ind.blocks:
            blocksize = block[1] - block[0]
            if blocksize > max_length_block:
                max_length_block = blocksize
        return max_length_block

    def apply(self, ind):
        # getting the block with the highest length
        max_length_block = self._get_block_maxsize(ind)
        # while at least one block is >than the preset max size
        while max_length_block > self.max_block_size:
            # choose a random block index
            bkpt = random.randint(0, len(ind.blocks) - 1)
            # evalaute his blocksize
            blocksize = ind.blocks[bkpt][1] - ind.blocks[bkpt][0]
            # if that block > than the preset max size
            if blocksize > self.max_block_size:
                # split the block according old_block = [min- max]
                # newblock = [min- random(min-max,min_block_size)][random(min-max,min_block_size)-max]
                second_end = ind.blocks[bkpt][1]
                first_end = random.randrange(
                    ind.blocks[bkpt][0] + self.min_block_size,
                    second_end - self.min_block_size, self.step_size)
                second_start = first_end
                ind.blocks[bkpt][1] = first_end
                newblock = [second_start, second_end]
                ind.blocks.insert((bkpt + 1), newblock)
            max_length_block = self._get_block_maxsize(ind)


"""
    Class SplitBlock

    desc
"""


class SplitBlockOperator(Operator):

    def __set_max_block_size(self):
        self.max_block_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.SplitBlockOperator"
        ]["max_block_size"]

    def __set_min_block_size(self):
        self.min_block_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.SplitBlockOperator"
        ]["min_block_size"]

    def __set_step_size(self):
        self.step_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.SplitBlockOperator"
        ]["step_size"]

    def initialise(self):
        self.__set_max_block_size()
        self.__set_min_block_size()
        self.__set_step_size()

    # split a block in two non overlapping
    def apply(self, ind):

        # making sure that blocks are sorted
        bkpts = list(range(0, len(ind.blocks)))
        bkpt = random.choice(bkpts)
        blocksize = ind.blocks[bkpt][1] - ind.blocks[bkpt][0]

        if blocksize >= 2 * self.min_block_size:
            second_end = ind.blocks[bkpt][1]

            try:
                first_end = random.randrange(ind.blocks[bkpt][0] + self.min_block_size,
                                             second_end - self.min_block_size, self.step_size)
            except ValueError:
                first_end = (ind.blocks[bkpt][0] + self.min_block_size)

            second_start = first_end
            second_end = ind.blocks[bkpt][1]
            ind.blocks[bkpt][1] = first_end
            newblock = [second_start, second_end]
            ind.blocks.insert((bkpt + 1), newblock)




"""
    Class JoinBlockOperator

    desc
"""


class JoinBlockOperator(Operator):

    def __set_max_block_size(self):
        self.max_block_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.JoinBlockOperator"
        ]["max_block_size"] - self.junction_size

    def __set_min_block_size(self):
        self.min_block_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.JoinBlockOperator"
        ]["min_block_size"]

    def __set_step_size(self):
        self.step_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.JoinBlockOperator"
        ]["step_size"]

    def __set_junction_size(self):
        self.junction_size = self.yaml["Algorithm"]["operators"][
            "mooda.operator.JoinBlockOperator"
        ]["junction_size"]


    def join(self, ind, block_1_pt, direction):
        # If forward join a block with next block
        if direction == "forward":
            block_2_pt = block_1_pt + 1
            ind.blocks[block_1_pt][1] = ind.blocks[block_2_pt][1]
            ind.blocks.remove(ind.blocks[block_2_pt])
            first_start = ind.blocks[block_1_pt][0]
            first_end = ind.blocks[block_1_pt][1]
            new_block_pt = block_1_pt


        # If backward join a block with previous
        elif direction == "backward":
            block_2_pt = block_1_pt - 1
            ind.blocks[block_2_pt][1] = ind.blocks[block_1_pt][1]
            ind.blocks.remove(ind.blocks[block_1_pt])
            first_start = ind.blocks[block_2_pt][0]
            first_end = ind.blocks[block_2_pt][1]
            new_block_pt = block_2_pt

        # if blocksize > max size split until  min<size<max
        block_size = first_end - first_start
        if first_end - first_start > self.max_block_size:
            break_range = math.floor(self.max_block_size - (block_size/2))
            median_point = math.floor((first_start + first_end)/2)
            if self.max_block_size - block_size > self.min_block_size:
                first_end = random.randrange((median_point - break_range),(median_point + break_range), self.step_size)
            else :
                first_end = median_point
            second_start = first_end
            second_end = ind.blocks[new_block_pt][1]
            ind.blocks[new_block_pt][1] = first_end
            newblock = [second_start, second_end]
            ind.blocks.insert((new_block_pt + 1), newblock)




            # while first_end - first_start > self.max_block_size or \
            #     second_end - second_start > (self.max_block_size + self.step_size):
            #     second_end = ind.blocks[new_block_pt][1]
            #     first_end = random.randrange(first_start + self.min_block_size,
            #                                  second_end - self.min_block_size, self.step_size)
            #     second_start = first_end
            #     second_end = ind.blocks[new_block_pt][1]



    def initialise(self):
        self.__set_junction_size()
        self.__set_step_size()
        self.__set_max_block_size()
        self.__set_min_block_size()


    def apply(self, ind):
        block_1_pt = random.randrange(0, len(ind.blocks) - 1)
        if block_1_pt == 0:
            direction = "forward"
            self.join(ind, block_1_pt, direction)
        elif 0 < block_1_pt < (len(ind.blocks) - 1):
            direction = random.choice(["forward", "backward"])
            self.join(ind, block_1_pt, direction)
        elif block_1_pt == (len(ind.blocks) - 1):
            direction = "forward"
            self.join(ind, block_1_pt, direction)


"""
    Class GCOptimizationOperator(Operator)
"""


class GCOptimizationOperator(Operator):
    def __set_codon_GC_table(self):
        self.codon_GC_table = CodonTable()
        self.codon_GC_table.codons = self.codon_GC_table.load_config(
            self.yaml["Algorithm"]["operators"]["mooda.operator.GCOptimizationOperator"][
                "codon_GC_table"
            ]
        )

    def initialise(self):
        self.__set_codon_GC_table()
        self.target_gc = self.yaml["Algorithm"]["operators"][
            "mooda.operator.GCOptimizationOperator"
        ]["target_gc"]
        self.step_size_gc = self.yaml["Algorithm"]["operators"][
            "mooda.operator.GCOptimizationOperator"
        ]["step_size"]

    def apply(self, ind):
        # reference sequence

        # picking a CDS at random
        cds_pt = random.randint(0, (len(ind.cds_list) - 1))
        # compute the GC of the CDS
        cds_gc = GC(ind.cds_list[cds_pt].seq)
        # get codons for the CDS
        # create and shuffle CDS indexes for hill climbing
        cds_codon_index = list(range(0, len(ind.cds_list[cds_pt].codons)))
        np.random.shuffle(cds_codon_index)

        # hill climbing GC mutator
        codon_it = 0
        accepted_step = False
        while not accepted_step and codon_it < len(ind.cds_list[cds_pt].codons):
            # picking a codon at random
            curr_codon = ind.cds_list[cds_pt].codons[cds_codon_index[codon_it]]
            # translate codon to aminoacid
            curr_aa = curr_codon.translate(table =ind.cds_list[cds_pt].translation_table)

            # get current codons and directionality
            # in a hill climbing fashion 333, 5, 4, 4,
            if cds_gc < self.target_gc:
                curr_aa_codons = self.codon_GC_table.get_codons(
                    str(curr_aa), order_asc_gc=False
                )
            elif cds_gc > self.target_gc:
                curr_aa_codons = self.codon_GC_table.get_codons(
                    str(curr_aa), order_asc_gc=False
                )
            else:
                curr_aa_codons = None

            # computing stop criterion
            if curr_aa_codons and (curr_aa_codons[0] != curr_codon):
                ind.cds_list[cds_pt].codons[cds_codon_index[codon_it]] = Seq(
                    curr_aa_codons[0], IUPAC.unambiguous_dna
                )
                # calculating GC just for the CDS.
                ind.cds_list[cds_pt].build_sequence_from_codons()

                accepted_step = (
                    np.abs(GC(ind.cds_list[cds_pt].seq) - cds_gc) >= self.step_size_gc
                )
            else:
                codon_it = codon_it + 1

            # updating individual sequence
        ind.cds_list[cds_pt].build_sequence_from_codons()
        ind.sequence = (
            ind.sequence[:ind.cds_list[cds_pt].pt.location.start]
            + ind.cds_list[cds_pt].seq
            + ind.sequence[ind.cds_list[cds_pt].pt.location.end:]
        )


"""
    Class CodonUsageOperator
"""


class CodonUsageOperator(Operator):
    def __set_codon_usage_table(self):
        self.codon_usage_table = CodonTable()
        self.codon_usage_table.codons = self.codon_usage_table.load_config(
            self.yaml["Algorithm"]["operators"]["mooda.operator.CodonUsageOperator"][
                "codon_usage_table"
            ]
        )

    def __set_step_size_codon_usage(self):
        self.step_size_codon_usage = (self.yaml["Algorithm"]["operators"]["mooda.operator.CodonUsageOperator"][
            "step_size"]) * 100

    def __swap_number_(self, codon_list):
        codons_to_swap = int(((len(codon_list)) * self.step_size_codon_usage) / 100)
        return codons_to_swap

    def initialise(self):
        self.__set_codon_usage_table()
        self.__set_step_size_codon_usage()

    def apply(self, ind):
        # reference sequence
        # checking that a CDS exists
        # picking a CDS at random
        cds_pt = random.randint(0, (len(ind.cds_list) - 1))
        # picking the CDS sequence
        # get codons for the CDS
        # number of indexes
        # evaluate the number of codons to edit
        codons_to_swap = self.__swap_number_(ind.cds_list[cds_pt].codons)
        codons_indexes = random.sample(
            range(0, (len(ind.cds_list[cds_pt].codons) - 1)), codons_to_swap
        )
        for index in codons_indexes:
            chosen_codon = ind.cds_list[cds_pt].codons[index]
            aa = chosen_codon.translate()
            new_codon = np.random.choice(
                list(self.codon_usage_table.codons[aa].keys()),
                p=list(self.codon_usage_table.codons[aa].values()),
            )
            new_codon = Seq(new_codon, IUPAC.unambiguous_dna)
            ind.cds_list[cds_pt].codons[index] = new_codon
        ind.cds_list[cds_pt].build_sequence_from_codons()
        ind.sequence = (
            ind.sequence[: ind.cds_list[cds_pt].pt.location.start]
            + ind.cds_list[cds_pt].seq
            + ind.sequence[ind.cds_list[cds_pt].pt.location.end:]
        )


"""
    Class Repetition
"""


class MotifOperator(Operator):

    def __set_repetition_table(self):
        self.repetition_table = RepetitionTable()
        self.repetition_table.motives_dict = self.repetition_table.load_config(
            self.yaml["Algorithm"]["operators"]["mooda.operator.MotifOperator"][
                "motif_table"
            ]
        )
        self.repetition_table.get_motives_list()

    def __set_codon_usage_table(self):
        self.codon_usage_table = CodonTable()
        self.codon_usage_table.codons = self.codon_usage_table.load_config(
            self.yaml["Algorithm"]["operators"]["mooda.operator.MotifOperator"][
                "codon_usage_table"
            ]
        )

    def __number_of_removal(self, repetition_list, cds):
        repetitions = 0
        for motive in repetition_list:
            n_rep = cds.count(motive)
            repetitions += n_rep
        removal_number = repetitions * self.step_size_rep
        return int(removal_number)

    def __remove_cds_without_rep(self, ind, motives_list):
        pt_list = []
        for motive in motives_list:
            for pt in range(0, len(ind.cds_list)):
                if pt not in pt_list and ind.cds_list[pt].seq.count(motive) != 0:
                    pt_list.append(pt)
        return pt_list

    def __select_repetion(self, repeats_list, ind, pt):
        motives = [
            motive for motive in repeats_list if ind.cds_list[pt].seq.count(motive) != 0
        ]
        return motives

    def __remove_motive(self, motive, ind, cds_pt):
        if ind.cds_list[cds_pt].seq.count(motive) > 0:
            cds_string = str(ind.cds_list[cds_pt].seq)
            start_rep = cds_string.index(motive)
            # getting the end index
            end_rep = start_rep + len(motive)
            # mapping this indexis on the list of codons
            start_codon = math.floor(start_rep / 3)
            end_codon = math.ceil(end_rep / 3) - 1
            # if the repetition is longer then 2 codon avoid the first and the last codon
            selected_codon_index = random.randint(start_codon, end_codon)
            selected_codon = ind.cds_list[cds_pt].codons[selected_codon_index]
            aa = selected_codon.translate()
            synonimous_codons = list(self.codon_usage_table.codons[aa].keys())
            if len(synonimous_codons) > 1:
                synonimous_codons.remove(selected_codon)
                new_codon = random.choice(synonimous_codons)
                new_codon = Seq(new_codon, IUPAC.unambiguous_dna)
                ind.cds_list[cds_pt].codons[selected_codon_index] = new_codon
                ind.cds_list[cds_pt].build_sequence_from_codons()

    def initialise(self):
        self.__set_repetition_table()
        self.__set_codon_usage_table()
        self.step_size_rep = (self.yaml["Algorithm"]["operators"]["mooda.operator.MotifOperator"]
        ["step_size"]) * 100

    def apply(self, ind):
        pts_motives_cds = self.__remove_cds_without_rep(
            ind, self.repetition_table.motives
        )
        if len(pts_motives_cds) > 0:
            cds_pt = random.choice(pts_motives_cds)
            to_remove = self.__select_repetion(
                self.repetition_table.motives, ind, cds_pt
            )
            op_iter = self.__number_of_removal(to_remove, ind.cds_list[cds_pt].seq)
            for ip_iter in range(op_iter):
                motive = random.choice(to_remove)
                self.__remove_motive(motive, ind, cds_pt)
            ind.sequence = (
                ind.sequence[: ind.cds_list[cds_pt].pt.location.start]
                + ind.cds_list[cds_pt].seq
                + ind.sequence[ind.cds_list[cds_pt].pt.location.end:]
            )



import logging
import argparse
import importlib
import sys

from mooda.objective_function import ObjectiveFunction
from mooda.operator import Operator
from mooda.assembly import Assembly
from mooda.io import MoodaFile
from mooda.evolutionary_algorithm import EvolutionaryAlgorithm
from mooda.montecarlo_algorithm import MonteCarloAlgorithm
from mooda.config import YamlConfig


def load_external_classes(config, section, class_type):
    cls_list = []

    for op, conf in config["Algorithm"][section].items():
        fields = op.split(".")

        # determining the name of the module and the class name
        # GS: You're trying to import the whole class name
        # GS: for example: to load splitblock operator, you have to pass something like
        # GS: mooda.Operator.SplitBlockOperator
        # GS: once the package is in your path, that's all you need.

        mod_name, cls_name = ".".join(fields[:-1]), fields[-1]

        logging.debug("Loading {} from {}".format(cls_name, mod_name))

        mod = importlib.import_module(mod_name)

        # tries to load and create an instance of the class
        try:
            cls = getattr(mod, cls_name)(config)
        except AttributeError:
            logging.error("Can not find class {}".format(cls))
            sys.exit(-1)

        # check if the class is of type Plugin
        if isinstance(cls, class_type):
            # exectute something
            cls.initialise()
            cls_list.append(cls)
        else:
            logging.error("{} is not a valid {}.".format(op, class_type))
            sys.exit(-1)
    return cls_list


##############################################################
###### CMD line options parser
##############################################################
def parse_command_line_options():
    parser = argparse.ArgumentParser(
        description="Multi Objective Optimisation for DNA design and Assembly."
    )
    parser.add_argument(
        "--input-file",
        "-i",
        type=str,
        required=True,
        help="A GenBank file containing the construct to redesign for assembly.",
    )

    parser.add_argument(
        "--algorithm",
        "-ag",
        type=str,
        choices=['mo', 'mc'],
        required=False,
        default='mo',
        help="Algorithm to use caa be either mo= Multi-Objective or mc= MonteCarlo.",
    )

    parser.add_argument(
        "--yaml-config",
        "-c",
        type=str,
        required=True,
        help="A .yaml configuration file describing all the parameters",
    )
    parser.add_argument(
        "--iterations",
        "-it",
        type=int,
        required=True,
        help="Algorithm number of iterations ",
    )
    parser.add_argument(
        "--population-size",
        "-p",
        type=int,
        required= False,
        help="number of individuals per iteration",
    )
    parser.add_argument(
        "--sequence_output",
        "-gf",
        type=bool,
        required=False,
        default=False,
        help="if true generates genbank and fasta files related to non"
        "dominated solutions",
    )
    parser.add_argument(
        "--archive-size",
        "-a",
        type=int,
        required=False,
        default=0,
        help="number of non dominated solutions stored at each iteration",
    )
    parser.add_argument(
        "--block-min-size", "-mns", type=int, required=True, help="block minimum size"
    )
    parser.add_argument(
        "--block-max-size", "-mxs", type=int, required=True, help="block maximum size"
    )
    parser.add_argument(
        "--block-step-size", "-bss", type=int, required=False, default=50, help="block maximum size"
    )
    parser.add_argument(
        "--block-junction-size", "-js", type=int, required=False, default=40, help="block junction size"
    )
    parser.add_argument(
        "--output-dir", "-dir", type=str, required=True, help="output directory"
    )
    options = parser.parse_args()
    return options


def main():
    # setting the logger for the whole section
    logging.basicConfig(level=logging.INFO)

    # parsing options
    options = parse_command_line_options()
    config = YamlConfig()
    config = config.load_config(options.yaml_config)

    # selecting the algorithm
    ag = None

    if options.algorithm == "mo":
        ag = EvolutionaryAlgorithm()
    elif options.algorithm == "mc":
        ag = MonteCarloAlgorithm()

    # loading settings from yaml and command line
    ag.load_setting(config, options)

    # setting operators
    ag.operators = load_external_classes(config, "operators", Operator)

    # setting objective functions
    ag.objective_functions = load_external_classes(
        config, "objective_functions", ObjectiveFunction
    )
    # setting assembly
    ag.assemblies = load_external_classes(
        config, "assemblies", Assembly
    )

    # running the algorithm
    ag.run()

    # saving files
    mf = MoodaFile(config)
    # GS: you don't need to pass config yaml again, it's already in the MoodaFile object
    mf.save(options, ag)


##############################################################
###### MAIN method
##############################################################
if __name__ == "__main__":
    main()



from Bio.Seq import Seq
from Bio.Alphabet import DNAAlphabet


class DNAComponent:
    def __init__(self):
        self.pt = None
        self.seq = None


class Coding(DNAComponent):
    def __init__(self):
        super().__init__()
        self.codons = None
        self.strand = None
        self.translation_table = None
        self.start_codon = None
        self.translation = None

    def __get_CDS(self, ind):
        self.seq = ind.sequence[self.pt.location.start : self.pt.location.end]
        if 'transl_table' in self.pt.qualifiers:
             self.translation_table = self.pt.qualifiers['transl_table'][0]
        else:
            self.translation_table = 1
        if 'codon_start' in self.pt.qualifiers:
            self.codon_start = self.pt.qualifiers['codon_start'][0]
        else:
            self.codon_start = 1
        if 'translation' in self.pt.qualifiers:
            self.translation = self.pt.qualifiers['translation'][0]

        if self.pt.location.strand != 1:
            self.seq = self.seq.complement()
            self.seq = self.seq[::-1]
        # print(self.seq)
        # print(self.pt)
        # print(self.pt.location)
        # print('table',self.translation_table)
        # a = self.seq.translate(table=self.translation_table, to_stop=True)
        # b = self.translation
        # print(a)
        # print(len(a), len(b))
        # if a==b:
        #     print("everythin ok")
        # else :
        #     print('trnaslation error')




    def __get_codon_list(self):
        self.codons = [self.seq[i : i + 3] for i in range(0, len(self.seq), 3)]

    def intialise_cds(self, ind, pt):
        self.pt = pt
        self.strand = self.pt.location.strand
        self.__get_CDS(ind)
        self.__get_codon_list()

    def build_sequence_from_codons(self):
        self.seq = sum(self.codons, Seq("", DNAAlphabet()))
        if self.strand != 1:
            self.seq = self.seq.complement()
            self.seq = self.seq[::-1]



class NonCoding(DNAComponent):
    def __get_UTR(self, ind):
        self.seq = ind.sequence[self.pt[0] : self.pt[1]]

    def intialise_utr(self, ind, pt):
        self.pt = pt
        self.__get_UTR(ind)




'''
    Class Individual

    Represents solution to a multi-objective DNA assembly problem

'''
import copy
from mooda.dna_component import Coding
from mooda.dna_component import NonCoding


class Individual:
    def __init__(self, gebank_record):
        # sequence information
        self.gebank_record = gebank_record
        self.sequence = self.gebank_record.seq
        self.features = self.gebank_record.features

        # list of blocks for assembly
        self.blocks = [[0, len(gebank_record.seq)]]
        # variables for the optimisation
        # objetctives
        self.objectives = []
        # rank of the solution
        self.rank = float("inf")
        # list of dominated solutions
        self.dominated_set = []
        # number of solutions that are dominated
        self.domination_count = 0
        # number of solutions that are better
        self.dominating_count = 0
        # crowding distance as defined by NSGAII
        self.crowding_distance = 0
        # list of CDS indexes included in the sequence
        self.cds_indexes_list = []
        # list of UTR indeces icluded in the sequence
        self.utr_indexes_list = []
        # list of codons
        self.cds_list = []
        # list of utr
        self.utr_list = []
        # fasta file
        self.fasta = None
        # genbank file
        self.genbank = None

    # representation of the solution
    def __str__(self):
        return str(self.objectives)

    def __eq__(self, other):
        if self.sequence == other.sequence \
            and self.blocks == other.blocks:
            return True
        else:
            return False

    # clone the instance with a deepcopy
    # useful when creating offsprings
    def clone(self):
        offspring = Individual(self.gebank_record)
        offspring.sequence = copy.deepcopy(self.sequence)
        offspring.features = copy.deepcopy(self.features)
        offspring.blocks = copy.deepcopy(self.blocks)
        offspring.cds_indexes_list = copy.deepcopy(self.cds_indexes_list)
        offspring.utr_indexes_list = copy.deepcopy(self.utr_indexes_list)
        offspring.cds_list = copy.deepcopy(self.cds_list)
        offspring.utr_list = copy.deepcopy(self.utr_list)
        return offspring

    def __get_feature_CDS_list(self):
        cds_index = []
        for feat in self.features:
            if feat.type == "CDS":
                cds_index.append(feat)
        self.cds_indexes_list = cds_index

    def __get_first_feature_UTR(self):
        if self.cds_indexes_list[0].location.start != 0:
            utr_pt = [0]
            utr_pt.append(self.cds_indexes_list[0].location.start - 1)
            self.utr_indexes_list.append(utr_pt)

    def __get_last_feature_UTR(self, pt):
        utr_pt = [None, None]
        utr_pt[0] = self.cds_indexes_list[pt].location.end + 1
        utr_pt[1] = len(self.sequence)
        if utr_pt[0] != utr_pt[1]:
            self.utr_indexes_list.append(utr_pt)

    def __get_feature_UTR_list(self):
        self.__get_first_feature_UTR()
        for pt in range(len(self.cds_indexes_list)):
            utr_pt = [None, None]
            if 0 <= pt < len(self.cds_indexes_list) - 1:
                utr_pt[0] = self.cds_indexes_list[pt].location.end + 1
                utr_pt[1] = self.cds_indexes_list[pt + 1].location.start - 1
                self.utr_indexes_list.append(utr_pt)
            elif pt == len(self.cds_indexes_list) - 1:
                self.__get_last_feature_UTR(pt)

    def __get_assembly(self, junction_size ):
        for block_index in range(len(self.blocks[:-1])):
            next_block = block_index + 1
            # if there is no overlap make it
            if self.blocks[block_index][1] == self.blocks[next_block][0]:
                self.blocks[block_index][1] = self.blocks[block_index][1] + junction_size

    def initialise(self):
        self.__get_feature_CDS_list()
        self.__get_feature_UTR_list()
        for pt in self.cds_indexes_list:
            cds = Coding()
            cds.intialise_cds(self, pt)
            self.cds_list.append(cds)
        for pt in self.utr_indexes_list:
            utr = NonCoding()
            utr.intialise_utr(self, pt)
            self.utr_list.append(utr)



"""
    Class EvolutionaryAlgorithm

    Our workhorse optimiser for the multi-objective DNA assembly problem.

"""
import logging
import numpy as np
import datetime
from Bio import SeqIO
from mooda.individual import Individual
from mooda.population import Population
from mooda.archive import Archive
from mooda.operator import BlockInitialiserOperator
import proglog



class EvolutionaryAlgorithm:
    def __init__(self):
        self.gebank_record = None
        self.population = Population()
        self.operators = []
        self.objective_functions = []
        self.assemblies = None
        self.current_iteration = 0
        self.max_iterations = None
        self.yaml = None
        self.dataframe = None
        self.archive = Archive()
        self.block_initialiser = BlockInitialiserOperator()
        self.time_start = None
        self.time_end = None
        self.outputfasta_gebank = None
        self.logger = proglog.default_bar_logger('bar')

    # load a genbank file assuming one record per file.

    def set_yaml_file(self, config):
        self.yaml = config

    def load_sequence(self, options):
        self.gebank_record = SeqIO.read(open(options.input_file, "r"), "genbank")

    def add_operator(self, op):
        self.operators.append(op)

    def add_objective_function(self, objf):
        self.objective_functions.append(objf)

    def add_assembly(self, assembly):
        self.assemblies = assembly

    def set_population_size(self, options):
        self.population.size = options.population_size

    def set_max_iterations(self, options):
        self.max_iterations = options.iterations

    def set_genbank_fasta_output(self, options):
        self.outputfasta_gebank = options.sequence_output

    def set_archive_length(self, options):
        self.archive.length = options.archive_size

    def set_block_initialiser(self, options):
        self.block_initialiser.step_size = options.block_step_size
        self.block_initialiser.max_block_size = options.block_max_size - options.block_junction_size
        self.block_initialiser.min_block_size = options.block_min_size

    def load_setting(self, config, options):
        self.set_yaml_file(config)
        self.load_sequence(options)
        self.set_population_size(options)
        self.set_max_iterations(options)
        self.set_genbank_fasta_output(options)
        self.set_archive_length(options)
        self.set_block_initialiser(options)

    def __initialize(self):
        for it in range(self.population.size):
            ind = Individual(self.gebank_record)
            ind.initialise()
            # evaluate objective functions for each individual
            self.block_initialiser.apply(ind)
            #self.assemblies[0].apply(ind)
            self.__eval_individual(ind)
            self.population.add_individual(ind)
        return self.population

    def run(self):
        self.time_start = datetime.datetime.now()
        self.__initialize()
        for it in self.logger.iter_bar(iteration=list(range(self.max_iterations))):
            self.current_iteration += 1
            #logging.info("itr=" + str(self.current_iteration))
            for i in range(len(self.population.individuals)):
                offspring = self.population.individuals[i].clone()
                op_curr = np.random.choice(self.operators)
                op_curr.apply(offspring)
                self.__eval_individual(offspring)
                self.population.individuals.append(offspring)
            self.population.compute_ranking()
            self.population.initialise_fronts()
            self.population.compute_crowding_distance(self.objective_functions)
            self.population.select_individuals()
            self.archive.add_to_archive(self.population)

            if len(self.archive.individuals) >= self.archive.length:
                self.archive.compute_ranking()
                self.archive.initialise_fronts()
                self.archive.select_only_rank1()
                self.archive.remove_duplicates()
                self.archive.random_deletion()

        self.assemblies[0].apply(self.population)
        self.__termination_process()
        self.time_end = datetime.datetime.now()
        duration = "duration= " + str(self.time_end - self.time_start)
        logging.info(duration)

    def __eval_individual(self, ind):
        ind.objectives = []
        # evaluate objfun for each individual
        for obj in self.objective_functions:
            ind.objectives.append(obj.eval(ind))


    def __termination_process(self):
        self.population.individuals = (
            self.population.individuals + self.archive.individuals
        )
        self.population.compute_ranking()
        self.population.initialise_fronts()
        self.population.compute_crowding_distance(self.objective_functions)
        self.population.select_only_rank1()
        self.population.remove_duplicates()
        self.population.individuals.sort(
            key=lambda i: i.crowding_distance, reverse=True
        )



"""
    Class ObjectiveFunction

    desc
"""
import numpy as np
from Bio.SeqUtils import GC
from mooda.config import CodonTable
from mooda.config import RepetitionTable


class ObjectiveFunction:
    def __init__(self, yaml):
        self.yaml = yaml


    def eval(self, ind):
        pass


"""
    Class GCContentObj

    desc
"""


class GCContentObjective(ObjectiveFunction):
    def __set_target_gc(self):
        self.target_gc = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.GCContentObjective"
        ]["target_gc"]

    def __set_junction_size(self):
        self.junction_size = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.GCContentObjective"
        ]["junction_size"]

    def initialise(self):
        self.__set_target_gc()
        self.__set_junction_size()

    def eval(self, ind):
        cum_sum = 0.0
        for pos in ind.blocks[:-1]:
            gc_val = GC(ind.sequence[pos[0]: (pos[1] + self.junction_size)])
            cum_sum += np.abs(gc_val - self.target_gc)

        pos = ind.blocks[-1]
        gc_val = GC(ind.sequence[pos[0]: pos[1]])
        cum_sum += np.abs(gc_val - self.target_gc)
        return cum_sum / float(len(ind.blocks))

    def __repr__(self):
        return 'GC content'


"""
    Class BlockVarianceObjective

    desc
"""


class BlockVarianceObjective(ObjectiveFunction):

    def __set_junction_size(self):
        self.junction_size = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.BlockVarianceObjective"
        ]["junction_size"]

    def initialise(self):
        self.__set_junction_size()

    def eval(self, ind):
        block_length_list = []

        for bb in ind.blocks[:-1]:
            blocksize = (bb[1] + self.junction_size) - bb[0]
            block_length_list.append(blocksize)

        bb = ind.blocks[-1]
        blocksize = bb[1] - bb[0]
        block_length_list.append(blocksize)

        block_variance = np.var(block_length_list, dtype=np.float64)
        return block_variance

    def __repr__(self):
        return 'Block variance'


class BlockNumberObjective(ObjectiveFunction):
    def initialise(self):
        pass

    def eval(self, ind):
        block_counter = len(ind.blocks)
        return block_counter

    def __repr__(self):
        return 'Block number'


"""
    Class BasePairCostObjective

    desc
"""


class BasePairCostObjective(ObjectiveFunction):

    def __set_junction_size(self):
        self.junction_size = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.BasePairCostObjective"
        ]["junction_size"]


    def __set_basepair_cost(self):
        self.basepair_cost = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.BasePairCostObjective"
        ]["basepair_cost"]

    def __set_block_cost(self):
        self.block_cost = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.BasePairCostObjective"
        ]["block_cost"]

    def initialise(self):
        self.__set_junction_size()
        self.__set_basepair_cost()
        self.__set_block_cost()

    def eval(self, ind):
        # turn blocks attribute into a list
        # for each block in the list
        cum_sum = 0.0

        for bb in ind.blocks[:-1]:
            blocksize = (bb[1] + self.junction_size) - bb[0]
            cum_sum += self.block_cost + blocksize * self.basepair_cost

        bb = ind.blocks[-1]
        blocksize = bb[1] - bb[0]
        cum_sum += self.block_cost + blocksize * self.basepair_cost
        return cum_sum

    def __repr__(self):
        return 'Cost'

"""
    Class CodonUsage

    desc
"""


class CodonUsageObjective(ObjectiveFunction):
    def set_codon_usage_table(self):
        self.codon_usage_table = CodonTable()
        self.codon_usage_table.codons = self.codon_usage_table.load_config(
            self.yaml["Algorithm"]["objective_functions"][
                "mooda.objective_function.CodonUsageObjective"
            ]["codon_usage_table"]
        )

    # command to return the highest frequency from
    def __get_codon_highest_frequency(self, aa):

        codon_dict = self.codon_usage_table.codons[aa]
        codon_highest_frequency = max(codon_dict, key=codon_dict.get)
        highest_frequency = codon_dict[codon_highest_frequency]
        return highest_frequency

    def __get_codon_frequency(self, codon):
        aa = codon.translate()
        codon_freq = self.codon_usage_table.codons[aa][codon]

        return codon_freq

    def initialise(self):
        self.set_codon_usage_table()

    def eval(self, ind):
        cum_sum = 0
        for index in ind.cds_indexes_list:

            #TODO: Trying to work with different genetic codes
            # for cds in ind.cds_list:
            #     if index == cds.pt:
            #         print(cds)


            cds_seq = ind.sequence[index.location.start:index.location.end]
            if index.location.strand != 1:
                cds_seq = cds_seq.complement()
                cds_seq = cds_seq[::-1]
            cds_codons = [cds_seq[i: i + 3] for i in range(0, len(cds_seq), 3)]
            for codon in cds_codons:
                aa = codon.translate()
                codon_frequence = self.__get_codon_frequency(codon)
                codon_frequence_target = self.__get_codon_highest_frequency(aa)
                cum_sum += abs(codon_frequence - codon_frequence_target)
        return cum_sum

    def __repr__(self):
        return 'Codon usage'



"""
    Class Repetition

    desc
"""


class MotifObjective(ObjectiveFunction):
    def initialise(self):
        self.repetition_table = RepetitionTable()
        self.repetition_table.motives_dict = self.repetition_table.load_config(
            self.yaml["Algorithm"]["objective_functions"][
                "mooda.objective_function.MotifObjective"
            ]["motif_table"]
        )
        self.repetition_table.get_motives_list()
        self.junction_size = self.yaml["Algorithm"]["objective_functions"][
            "mooda.objective_function.MotifObjective"
        ]["junction_size"]

    def eval(self, ind):
        # objective Function value
        cum_sum = 0.0
        # for each CDS in
        for pos in ind.blocks[:-1]:
            block_sequence = ind.sequence[pos[0]: pos[1] + self.junction_size]
            for motive in self.repetition_table.motives:
                cum_sum += block_sequence.count(motive)

        pos = ind.blocks[-1]
        block_sequence = ind.sequence[pos[0]: pos[1]]
        for motive in self.repetition_table.motives:
            cum_sum += block_sequence.count(motive)
        return cum_sum

    def __repr__(self):
        return 'Motifs & Repeats'



"""
    Class EvolutionaryAlgorithm

    Our workhorse optimiser for the multi-objective DNA assembly problem.

"""
import logging
import numpy as np
import datetime
import random
from Bio import SeqIO
from mooda.individual import Individual
from mooda.archive import Archive
from mooda.operator import BlockInitialiserOperator


class MonteCarloAlgorithm:
    def __init__(self):
        self.gebank_record = None
        self.individual = None
        self.operators = []
        self.objective_functions = []
        self.assemblies = None
        self.current_iteration = 0
        self.max_iterations = None
        self.yaml = None
        self.dataframe = None
        self.archive = Archive()
        self.block_initialiser = BlockInitialiserOperator()
        self.time_start = None
        self.time_end = None
        self.outputfasta_gebank = None

    # load a genbank file assuming one record per file.

    def set_yaml_file(self, config):
        self.yaml = config

    def load_sequence(self, options):
        self.gebank_record = SeqIO.read(open(options.input_file, "r"), "genbank")

    def add_operator(self, op):
        self.operators.append(op)

    def add_objective_function(self, objf):
        self.objective_functions.append(objf)

    def set_max_iterations(self, options):
        self.max_iterations = options.iterations

    def set_genbank_fasta_output(self, options):
        self.outputfasta_gebank = options.sequence_output

    def set_archive_length(self, options):
        self.archive.length = options.archive_size

    def set_block_initialiser(self, options):
        self.block_initialiser.step_size = options.block_step_size
        self.block_initialiser.max_block_size = options.block_max_size - options.block_junction_size
        self.block_initialiser.min_block_size = options.block_min_size

    def load_setting(self, config, options):
        self.set_yaml_file(config)
        self.load_sequence(options)
        self.set_max_iterations(options)
        self.set_genbank_fasta_output(options)
        self.set_archive_length(options)
        self.set_block_initialiser(options)

    def __initialize(self):
        self.individual = Individual(self.gebank_record)
        self.individual.initialise()
        # evaluate objective functions for each individual
        self.block_initialiser.apply(self.individual)
        self.__eval_individual(self.individual)

    def run(self):
        self.time_start = datetime.datetime.now()
        self.__initialize()
        for it in range(self.max_iterations):
            self.current_iteration += 1
            logging.info("itr=" + str(self.current_iteration))
            offspring = self.individual.clone()
            op_curr = np.random.choice(self.operators)
            op_curr.apply(offspring)
            self.__eval_individual(offspring)
            self.__duel(self.individual, offspring)
        self.__termination()
        self.time_end = datetime.datetime.now()
        duration = "duration= " + str(self.time_end - self.time_start)
        logging.info(duration)

    def __duel(self, parent, offspring):
        ind_domination_counter = 0
        offspring_domination_counter = 0
        for objective in range(0, (len(parent.objectives))):
            if parent.objectives[objective] < offspring.objectives[objective]:
                ind_domination_counter += 1
        for objective in range(0, (len(parent.objectives))):
            if offspring.objectives[objective] < parent.objectives[objective]:
                offspring_domination_counter += 1
        if offspring_domination_counter > ind_domination_counter:
            self.individual = offspring
            self.archive.individuals.append(offspring)
        elif offspring_domination_counter == ind_domination_counter:
            self.archive.individuals.append(offspring)
            self.individual = random.choice([offspring, parent])

    def __termination(self):
        self.archive.compute_ranking()
        self.archive.initialise_fronts()
        self.archive.compute_crowding_distance(self.objective_functions)
        self.archive.select_only_rank1()
        self.archive.remove_duplicates()
        self.assemblies[0].apply(self.archive)



    def __eval_individual(self, ind):
        ind.objectives = []
        # evaluate objfun for each individual
        for obj in self.objective_functions:
            ind.objectives.append(obj.eval(ind))



"""
    Class CodonUsageTable

    desc
"""
import yaml
from Bio.SeqUtils import GC


class YamlConfig:
    def __init__(self):
        pass

    def load_config(self, filename):
        with open(filename, "r") as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)
            return config


class CodonTable(YamlConfig):
    def __init__(self):
        super().__init__()
        self.codons = dict()

    def get_codons(self, aa, order_asc_gc=True):
        if order_asc_gc:
            return sorted(self.codons[aa], key=lambda k: GC(k))
        else:
            return sorted(self.codons[aa], key=lambda k: -GC(k))

    def get_codon_probabilities(self, aa):
        return self.codons[aa]


class RepetitionTable(YamlConfig):
    def __init__(self):
        super().__init__()
        self.motives_dict = dict()
        self.motives = []

    def get_repetition_len(self, rep):
        return self.motives_dict[rep]

    def get_motives_list(self):
        for unit in self.motives_dict:
            whole_motive = unit * self.motives_dict[unit]
            self.motives.append(whole_motive)




import os
from setuptools import find_packages, setup, Extension

# determining the directory containing setup.py
setup_path = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(setup_path, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

setup(
    # package information
    name = 'mooda-dna',
    packages = find_packages(),
    version = '0.7.0-dev',
    description = 'A Multi-Objective algorithm for DNA Design and Assembly',
    long_description = readme,
    long_description_content_type="text/markdown",
    license = 'MIT',
    url='https://github.com/stracquadaniolab/mooda',
    keywords='',

    # author information
    author = 'Angelo Gaeta, Giovanni Stracquadanio',
    author_email = 'a.gaeta@sms.ed.ac.uk',

    # installation info and requirements
    install_requires=[
        "biopython==1.73",
        "numpy==1.16.2",
        "pandas==0.24.2",
        "PyYAML==5.1",
        "proglog",
    ],
    setup_requires=[],

    # test info and requirements
    test_suite='tests',
    tests_require=[

    ],

    # package deployment info
    include_package_data=True,
    zip_safe=False,

    # all tools have cli interface
    entry_points={
        'console_scripts': [
            'mooda=mooda.cli:main',
        ],
    },

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ]
)



'''
    Class Archive

    Represents a list of all non-dominated solutions for a multi-objective DNA assembly problem

'''


class Archive:
    def __init__(self):
        # instance variabe with list of solutions
        self.individuals = []

        # list of non-dominated fronts
        self.fronts = [[]]
        # Number of individuals allowed in the archive
        self.length = None

    def __len__(self):
        return len(self.individuals)

    def add_individual(self, ind):
        self.individuals.append(ind)

    def add_to_archive(self, population):
        # add to archive
        for individual in population.fronts[0]:
            if individual not in self.individuals:
                self.individuals.append(individual)

    def compute_ranking(self):
        self.fronts = [[]]
        for p in self.individuals:

            p.dominated_set = []
            p.dominating_count = 0
            for l in self.individuals:
                p_counter = 0
                for objective in range(0, (len(p.objectives))):
                    if p.objectives[objective] < l.objectives[objective]:
                        p_counter += 1
                    elif p.objectives[objective] > l.objectives[objective]:
                        p_counter = 0
                        break
                if p_counter >= 1:
                    p.dominated_set.append(l)
                l_counter = 0
                for objective in range(0, len(p.objectives)):
                    if l.objectives[objective] < p.objectives[objective]:
                        l_counter += 1
                    elif l.objectives[objective] > p.objectives[objective]:
                        l_counter = 0
                        break
                if l_counter >= 1:
                    p.dominating_count += 1
            if p.dominating_count == 0:
                p.rank = 1
                self.fronts[0].append(p)

    def initialise_fronts(self):
        counter = 1
        while len(self.fronts[counter - 1]) > 0:
            next_front = []
            for p in self.fronts[counter - 1]:
                for l in p.dominated_set:
                    l.dominating_count -= 1
                    if l.dominating_count == 0:
                        l.rank = counter + 1
                        next_front.append(l)
            counter += 1
            self.fronts.append(next_front)

    def compute_crowding_distance(self, objective_functions):
        if len(self.individuals) > 2:
            for individual in self.individuals:
                individual.crowding_distance = 0

            for obj in range(0, len(objective_functions)):
                self.individuals.sort(key=lambda individual: individual.objectives[obj])
                min = self.individuals[0].objectives[obj]
                max = self.individuals[-1].objectives[obj]
                for front in self.fronts:
                    if len(front) > 2:
                        front.sort(key=lambda individual: individual.objectives[obj])
                        front[0].crowding_distance = front[-1].crowding_distance = 1
                        for individual_index in range(1, len(front) - 1):
                            try:
                                front[individual_index].crowding_distance += abs(
                                    front[individual_index - 1].objectives[obj]
                                    - front[individual_index + 1].objectives[obj]
                                ) / (abs(min - max))
                            except ZeroDivisionError:
                                front[individual_index].crowding_distance += 0

    def remove_duplicates(self):
        new_individuals = []
        for ind in self.individuals:
            if ind not in new_individuals:
                new_individuals.append(ind)
        self.individuals = new_individuals

    def select_only_rank1(self):
        self.individuals = self.fronts[0]

    def random_deletion(self):
        if len(self.individuals) >= self.length:
            self.individuals.sort(
                key=lambda individual: (individual.rank, -individual.crowding_distance))
        while len(self.individuals) > self.length:
            del self.individuals[-1]
